# encoding=CP1251
import nltk, os, math
import cPickle

numberUnigrams = 3
numberBigrams = 12
pathToCollection = "./open/"
pathToModel = "./models/ngram.pkl"


# all methods for text processing
class TextProcessor():
    def __init__(self, text):
        words = nltk.PunktWordTokenizer().tokenize(text)
        words = filter(lambda x: len(x) > 3, words)  # filter 1-2 symbols words
        stemmer = nltk.stem.snowball.RussianStemmer() 
        self.words = map(stemmer.stem, words)  # apply stemmer to each word
        self.originalForm = self.saveOriginalFormOfWords(words)
    
    # save original form of word
    # here it's first occurred word of the text corresponded to the particular token
    def saveOriginalFormOfWords(self, tokenizedText):            
        originalForm = dict()
        for i in xrange(len(self.words)):
            if not(originalForm.has_key(self.words[i])):
                originalForm[self.words[i]] = tokenizedText[i]
        return originalForm  
    
    # simple word restorer
    def wordRestorer(self, word):
        return self.originalForm[word]
        
# training of algorithms
# here we only compute IDf
class Trainer():
    # model is a pair of (dict: word -> idf_score) and (number of files in training set)
    model = None
    
    def __init__ (self):
        if not(self.loadModel()):
            self.model = self.computeIDF()
            self.saveModel()
        
    def computeIDF(self):
        fileCounter = 0
        docs = dict() 
        for fileName in os.listdir(pathToCollection):
            print fileCounter,  # progress
            fileCounter += 1
                        
            f = open(pathToCollection + fileName)
            text = unicode(f.read(), "utf-8")
            f.close()
            
            # preprocess text
            tp = TextProcessor(text) 
            
            # process unigrams
            for unigram in set(tp.words):
                docs[unigram] = docs.setdefault(unigram, 0.0) + 1.0
            
            # process bigrams
            for bigram in set(NgramProcessor().getBigrams(tp.words)):
                docs[bigram] = docs.setdefault(bigram, 0.0) + 1.0
                
            
        # compute IDF
        idf = dict()
        for ngram in docs.keys():
            if docs[ngram] > 1:
                idf[ngram] = -math.log(docs[ngram] / fileCounter)
            
        return (idf, fileCounter)
    
    def saveModel(self):
        f = open(pathToModel, "wb")
        cPickle.dump(self.model, f)
        f.close()
    
    def loadModel(self):
        try:
            f = open(pathToModel)
            self.model = cPickle.load(f)
            f.close()
            return True
        except:
            return False
        
    def getModel(self):
        return self.model

# all functions for processing n-gramms
class NgramProcessor():
    def getBigrams(self, wordList):
        return zip(wordList[:-1], wordList[1:])
        
    # return sorted list of pairs (object, number of its occurrences in the list)
    # it is used for counting unigrams and bigrams in text
    def objectCounter(self, listOfObjects):
        return [(obj, listOfObjects.count(obj)) for obj in set(listOfObjects)] 
        
    # return TopN object (unigram of bigram) from list    
    def getTopN(self, listOfObjectsWithCounts, N):
        sortedListOfObjectsWithCounts = sorted(listOfObjectsWithCounts, key=lambda x:-int(x[1]))[:N]
        # return sortedListOfObjectsWithCounts
        return [item[0] for item in sortedListOfObjectsWithCounts]    

    # Assign weights to unigrams and bigrams
    # here we use TF*IDF weighting scheme
    def weightingScheme(self, listOfCountedObjects, model):
        return [(objectName, objectCount * model[0].setdefault(objectName, -math.log(1.0 / model[1]))) 
                for objectName, objectCount in listOfCountedObjects]
            
class Keywords():
    
    def __init__(self):
        # load trained model
        self.model = Trainer().getModel()
        
    def getKeywords(self, text):
        # text processor
        tp = TextProcessor(text)

        ng = NgramProcessor()
        bigrams = ng.getBigrams(tp.words)

        unigramKeywords = map(tp.wordRestorer, #restore original forms
                              ng.getTopN(ng.weightingScheme(ng.objectCounter(tp.words), self.model), 
                                         numberUnigrams))
        bigramKeywords = [tp.wordRestorer(word[0]) + " " + tp.wordRestorer(word[1]) 
                          for word in ng.getTopN(ng.weightingScheme(ng.objectCounter(bigrams),
                                                                  self.model), numberBigrams)]
        
        return unigramKeywords + bigramKeywords
        

## -=test=-
#kw = Keywords()
#f = open("./open/isp_20_2011_283.txt")
#text = unicode(f.read(), "utf-8")
#f.close()
#keywords = kw.getKeywords(text)
#print repr(keywords).decode("unicode_escape").encode("cp1251")
